from functools import wraps

from .core import QUERY_WRAPPED_METHOD
from .locators import Locator
from .injector import Injector


def deps(init_method):
    """
     - The `@deps` decorator on `__init__` enables _private injection_.
     - The special method `__deps__` define which dependencies are required.
     - All the dependencies are located and injected before the object is initialized.
     - Finally, the real `__init__` method is called.

    Notice that `__deps__` is a mock method, it is not ever called.
    Its sole purpose is to define the dependencies through its argument signature.
    """
    assert init_method.__name__ == '__init__', \
        '"{}" decorator must precede __init__ method.'.format(deps.__name__)

    @wraps(init_method)
    def init_wrapper(self, *args, _dependencies=None, **kwargs):
        # Query wrapped init
        if self == QUERY_WRAPPED_METHOD:
            return init_method

        # Inject
        if _dependencies:
            for name, value in _dependencies.items():
                setattr(self, name, value)

        # run __init__
        init_method(self, *args, **kwargs)

    return init_wrapper


def inject(*dependencies: Locator,
           injector: Injector = None
           ):
    """
    @inject inspects the arguments of the wrapped function signature and inject the required dependencies.
    You get all the needed dependencies through the function arguments.

    It takes one or more Locator objects as arguments such as Dependencies or Autowiring. They will used to create an
    Injector object which will be used to perform the injection process.

    Alternative, you can pass directly an Injector object through the named argument "injector". In this case,
    do not pass Locator objects.

    """
    if injector:
        assert not dependencies, \
            '"{}" decorator requires one argument only, two provided.'.format(inject.__name__)
        inj = injector
    elif dependencies:
        inj = Injector(*dependencies)
    else:
        raise AssertionError('"{0}" decorator requires an {1} or some {1}s, nothing provided.'
                             .format(inject.__name__, Injector.__name__, Locator.__name__))

    def get_func_wrapper(func):

        def func_wrapper(*args, **kwargs):
            return inj.call(func, *args, **kwargs)

        return func_wrapper

    return get_func_wrapper



from wirinj import Singleton, inject, Definitions


class Pet:
    pass


class Cat(Pet):
    pass


class Dog(Pet):
    pass


defs = {
    'cat': Singleton(Cat),
    Pet: Singleton(Dog),
}


@inject(Definitions(defs))
def fn(cat, pet: Pet):
    print('cat is a', cat.__class__.__name__)
    print('pet is a', pet.__class__.__name__)


fn()



import logging
from io import StringIO
from typing import Type

from unittest import TestCase

from wirinj.decorators import inject, deps
from wirinj import Autowiring, Definitions

from examples.pet_delivery.classes import Mike, Engine, Pet
from examples.pet_delivery.defs import world_one


class Reality:
    pass


class Thing:

    def __deps__(self, reality: Reality):
        pass

    @deps
    def __init__(self, param):
        self.param = param


class Obj:
    def __deps__(self, dependency):
        #This is never executed. Makes work IDE inspections for "dependency".
        self.dependency = dependency

    @deps
    def __init__(self, param):
        self.param = param


class TestDeps(TestCase):

    def test_dependencies_argument(self):
        obj = Obj(10, _dependencies={'dependency': 20})
        self.assertEqual(obj.dependency, 20)
        self.assertEqual(obj.param, 10)

    def test_factory_with_arguments(self):

        @inject(Autowiring())
        def fn(factory: Type[Thing]):
            thing = factory(10)
            self.assertIsInstance(thing, Thing)
            self.assertIsInstance(thing.reality, Reality)
        fn()

    def test_mixed_priv_and_public_args(self):

        @inject(Definitions({'param': 10}), Autowiring())
        def fn(thing: Thing):
            self.assertIsInstance(thing, Thing)
            self.assertIsInstance(thing.reality, Reality)
            self.assertIs(thing.param, 10)
        fn()

    def test_deps_on_class_and_subclass(self):
        class Qux:
            pass

        class Foo:
            pass

        class Bar:

            def __deps__(self, foo: Foo, foo2: Foo, **_):
                pass

            @deps
            def __init__(self):
                pass

        class Baz(Bar):

            def __deps__(self, qux: Qux, **_):
                pass

            @deps
            def __init__(self):
                super().__init__()

        @inject(Autowiring())
        def do(baz: Baz):
            self.assertIsInstance(baz.foo, Foo)
            self.assertIsInstance(baz.foo2, Foo)
            self.assertIsInstance(baz.qux, Qux)
            self.assertIs(baz.foo, baz.foo2)
        do()



class TestInject(TestCase):
    def test_inject_full_pet_delivery_example(self):
        log = StringIO()
        logging.basicConfig(stream=log, level=logging.INFO, format='%(message)s')

        @inject(world_one)
        def do(mike: Mike):
            mike.deliver(100, 5, False)

        do()

        log_str = log.getvalue()
        self.assertMultiLineEqual(
            log_str,
            'Picking pets up:  MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\n',
            'Mismatch in output of test_classes',
        )



from wirinj import inject, Definitions, Instance


class Animal:
    def __init__(self, sound):
        self.sound = sound


class Dog(Animal):
    pass


class Cat(Animal):
    pass


class Cow(Animal):
    pass


defs = {
    Dog: Instance(),
    Cat: Instance(),
    Cow: Instance(),

    (Dog, 'sound'): 'woof',
    (Cat, 'sound'): 'meow',
    'sound': '?',
}


@inject(Definitions(defs))
def fn(cat: Cat, dog: Dog, cow: Cow):
    print('Cat:', cat.sound)
    print('Dog:', dog.sound)
    print('Cow:', cow.sound)


fn()



from random import randint

from wirinj import CustomInstance, inject, Definitions


class Cat:
    def __init__(self, color, weight):
        self.color = color
        self.weight = weight

    def __str__(self):
        return f'A {self.color} pounds {self.weight} cat.'


def create_cat(color):
    return Cat(color, randint(4, 20))


defs = {
    'color': 'blue',
    Cat: CustomInstance(create_cat),
}


@inject(Definitions(defs))
def fn(cat1: Cat, cat2: Cat, cat3: Cat):
    print(cat1)
    print(cat2)
    print(cat3)


fn()



from typing import Type


def get_subclassing_factory(cls, func):
    base_cls = cls

    class FactoryMeta(cls.__class__):
        def __call__(cls, *args, **kwargs):
            return func(base_cls, *args, **kwargs)

    class Factory(cls, metaclass=FactoryMeta):
        pass

    return Factory


def get_func_factory(cls, func):
    base_cls = cls

    def factory(*args, **kwargs):
        return func(base_cls, *args, **kwargs)

    return factory


def is_typing_type(cls):
    return cls.__class__ is Type.__class__ and getattr(cls, '_name', getattr(cls, '__name__', None)) == 'Type'

def is_typing_clause(cls):
    return cls.__module__ == 'typing'


def get_typing_args(cls):
    return cls.__args__[0] if cls.__args__ else None


def get_cls_name(cls):
    if is_typing_type(cls):
        args = get_typing_args(cls)
        if args is None:
            a = 1
        return 'Type[{}]'.format(args.__name__)
    else:
        return cls.__name__



from typing import Any
from unittest import TestCase

from wirinj import INJECTED
from wirinj.injector import Injector
from wirinj.definition import Definitions, CustomSingleton, Singleton, CustomInstance
from wirinj.decorators import inject

class TestCustomSingletonDependency(TestCase):

    def test_full_injection(self):
        class Foo:
            def __init__(self, bar):
                self.bar = bar

        def foo_creator(bar) -> Foo:
            return Foo(bar)

        deps = Definitions({
            'bar': 'Hola',
            Foo: CustomSingleton(foo_creator),
        })

        @inject(deps)
        def do(foo: Foo):
            self.assertIsNotNone(foo)
            self.assertIsNotNone(foo.bar)

        do()

class TestCustomInstanceDependency(TestCase):

    def test_custom_instance_with_params(self):
        class Baz:
            pass

        class Foo:
            def __init__(self, bar, baz: Baz = INJECTED):
                self.bar = bar
                self.baz = baz

        def foo_creator(bar, baz: Baz):
            return Foo(bar, baz)

        injector = Injector(Definitions({
            Baz: Singleton(),
            Foo: CustomInstance(foo_creator),
        }))

        foo = injector.get(Foo, 'Ping')
        self.assertIs(foo.bar, 'Ping')
        self.assertIsNotNone(foo.baz)

        foo2 = injector.get(Foo, bar='Pong')
        self.assertIs(foo2.bar, 'Pong')
        self.assertIsNotNone(foo2.baz)

        pass



from typing import Type

from wirinj import inject, Autowiring


class Cat:
    def __init__(self, sound):
        pass


@inject(Autowiring())
def fn(cat_factory: Type[Cat]):
    cat = cat_factory('Meow')
    print('cat:', cat)


fn()



from typing import Type

from wirinj import inject, Autowiring, deps, INJECTED


class Feeder:
    pass


class Cat:
    feeder: Feeder = INJECTED

    def __init__(self, color, weight):
        self.color = color
        self.weight = weight


@inject(Autowiring())
def fn(factory: Type[Cat]):
    cat = factory('blue', 12)
    print('cat.feeder:', cat.feeder)


fn()



from random import randint
from typing import Type

from wirinj import CustomInstance, Factory, inject, Definitions


class Cat:
    def __init__(self, name, color=None, weight=None):
        self.name = name
        self.color = color
        self.weight = weight

    def __str__(self):
        return f'{self.name} is a {self.color} pounds {self.weight} cat.'


def create_cat(name, color):
    return Cat(name, color, randint(4, 20))


defs = {
    'color': 'black',
    Cat: CustomInstance(create_cat),
    Type[Cat]: Factory(),
}


@inject(Definitions(defs))
def fn(factory: Type[Cat]):
    cat = factory('Tom')
    print(cat)
    cat2 = factory('Sam')
    print(cat2)


fn()



from typing import Type

from wirinj import inject, Autowiring, INJECTED


class Feeder:
    pass


class Cat:
    def __init__(self, color, weight, feeder: Feeder = INJECTED):
        self.color = color
        self.weight = weight
        self.feeder = feeder


@inject(Autowiring())
def fn(factory: Type[Cat]):
    cat = factory('blue', 12)
    print('cat.feeder:', cat.feeder)


fn()



from typing import Type

from wirinj import Definitions, Instance, Singleton, Factory, INJECTED, inject


class MyService:
    def __str__(self):
        return f'<{self.__class__.__name__}>'


class MyObject:
    my_service: MyService = INJECTED
    my_config: str = INJECTED

    def __str__(self):
        return f'<{self.__class__.__name__}> -> my_config: "{self.my_config}", param: {self.param}, my_service: {self.my_service}'

    def __init__(self, param):
        self.param = param


# Wiring definitions
defs = {
    'my_config': 'Some config',
    MyService: Singleton(),
    MyObject: Instance(),
    Type[MyObject]: Factory(),
}


# The decorator will inject a dependency on each function parameter based on its name and/or annotation type
@inject(Definitions(defs))
def do(
        my_service: MyService,
        my_object_factory: Type[MyObject]
):
    print('my_service = {}'.format(my_service))

    my_object1 = my_object_factory(10)
    print('my_object1 = {}'.format(my_object1))


# Inject and run it
do()



from typing import Type
from wirinj import Instance, inject, Definitions, Autowiring, AutowiringReport


class Cat:
    pass


class Dog:
    pass


class Horse:
    pass


deps = {
    Cat: Instance(),
}

report = AutowiringReport()


@inject(Definitions(deps), Autowiring(report))
def fn(cat: Cat, dog: Dog, horse_factory: Type[Horse]):
    print(cat.__class__.__name__)
    print(dog.__class__.__name__)
    horse = horse_factory()
    print(horse.__class__.__name__)


fn()

print(report.get())



from random import randint
from typing import Type

from wirinj import Singleton, Instance, Factory, CustomInstance, inject, Definitions


class House:
    pass


class Cat:
    pass


class Dog:
    pass


def dog_builder():
    """ Custom instantiation """
    dog = Dog()
    dog.random = randint(50, 100)
    return dog


defs = {
    House: Singleton(),
    Cat: Instance(),
    Type[Cat]: Factory(),
    Dog: CustomInstance(dog_builder),
    Type[Dog]: Factory(),
}


@inject(Definitions(defs))
def fn(house: House, cat_factory: Type[Cat], dog_factory: Type[Dog]):
    cat = cat_factory()
    dog = dog_factory()

    print('house:', house)
    print('cat:', cat)
    print('dog:', dog)


fn()



from examples.report.cat_example_classes import Nail, Cat
from wirinj import Injector, Definitions, Autowiring, Instance


class BrokenNail(Nail):

    def __init__(self):
        raise AttributeError('Broken leg')


inj = Injector(
    Definitions({
        Nail: Instance(BrokenNail)
    }),
    Autowiring(use_singletons=False),
)

inj.get(Cat)



from logging import getLogger
from typing import Type, Dict, List, TypeVar, Any

from wirinj import deps, INJECTED

logger = getLogger(__name__)


class Pet(object):
    sound: str = INJECTED
    weight: float = INJECTED

    def __init__(self, gift_wrapped):
        self.gift_wrapped = gift_wrapped

    def cry(self):
        return self.sound.lower() if self.gift_wrapped else self.sound.upper()


class Cat(Pet):
    pass


class Dog(Pet):
    pass


class Bird(Pet):
    pass


class Part:
    mount_sound: str = INJECTED

    def mount(self):
        return self.mount_sound


class Engine(Part):
    pass


class Plate(Part):
    pass


class Wheel(Part):
    pass


class Container(Part):
    pass


class VehicleBuilder:

    engine_factory: Type[Engine] = INJECTED
    plate_factory: Type[Plate] = INJECTED
    wheel_factory: Type[Wheel] = INJECTED
    container_factory: Type[Container] = INJECTED

    def build(self, recipe: Dict):
        parts = []  # type: List[Part]
        parts += [self.engine_factory() for _ in range(recipe.get('engines', 0))]
        parts += [self.plate_factory() for _ in range(recipe.get('plates', 0))]
        parts += [self.wheel_factory() for _ in range(recipe.get('wheels', 0))]
        parts += [self.container_factory() for _ in range(recipe.get('containers', 0))]

        mounting = ''
        for part in parts:
            mounting += ' ' + part.mount()

        return mounting


class Vehicle:
    builder: VehicleBuilder = INJECTED
    recipe: Dict = INJECTED
    max_load_weight: float = INJECTED

    def __init__(self):
        self.pets = []
        self.build()

    def go(self, miles):
        logger.info('{} goes {} miles'.format(self.__class__.__name__, miles))

    def come_back(self):
        logger.info('{} commes back'.format(self.__class__.__name__))

    def build(self):
        logger.info('{} is built: {}'.format(
            self.__class__.__name__,
            self.builder.build(self.recipe)
        ))

    def get_available_load(self):
        return self.max_load_weight - sum(pet.weight for pet in self.pets)


class Car(Vehicle):
    pass


class Van(Vehicle):
    pass


class Truck(Vehicle):
    pass


class PetLoader:

    def upload(self, pets: List[Pet], vehicle: Vehicle):
        info = 'Uploading to the {}:'.format(vehicle.__class__.__name__.lower())
        while pets:
            pet = pets.pop()
            if vehicle.get_available_load() >= pet.weight:
                vehicle.pets.append(pet)
                info += ' ' + pet.__class__.__name__
            else:
                pets.append(pet)
                break
        logger.info(info)

    def download(self, vehicle):
        logger.info('{} pets delivered'.format(len(vehicle.pets)))
        vehicle.pets = []


class PetPicker:

    pet_store: Type[Pet] = INJECTED

    def pick(self, qty, gift_wrapped):
        info = 'Picking pets up: '
        pets = []
        for _ in range(qty):
            pet = self.pet_store(gift_wrapped)
            info += ' ' + pet.cry()
            pets.append(pet)
        logger.info(info)
        return pets


class PetDeliveryPerson:

    def deliver(self, pet_qty, miles, gift_wrapped):
        pass


class Bob(PetDeliveryPerson):
    """Bob builds a car and deliver pets in his vehicle repeating the route several times."""

    vehicle: Vehicle = INJECTED
    pet_picker: PetPicker = INJECTED
    pet_loader: PetLoader = INJECTED

    def deliver(self, pet_qty, miles, gift_wrapped):
        # Pick up pets
        pets = self.pet_picker.pick(pet_qty, gift_wrapped)

        # Bob owns one vehicle only
        while pets:
            self.pet_loader.upload(pets, self.vehicle)
            self.vehicle.go(miles)
            self.pet_loader.download(self.vehicle)
            self.vehicle.come_back()


class Mike(PetDeliveryPerson):
    """Mike builds several autonomous vehicles and use them to deliver the pets all together"""

    vehicle_factory: Type[Vehicle] = INJECTED
    pet_picker: PetPicker = INJECTED
    pet_loader: PetLoader = INJECTED

    vehicles: List[Vehicle] = []

    def get_vehicle(self):
        if self.vehicles:
            return self.vehicles.pop()
        else:
            return self.vehicle_factory()

    def park_vehicles(self, vehicles):
        self.vehicles += vehicles

    def deliver(self, pet_qty, miles, gift_wrapped):

        # Pick up pets
        pets = self.pet_picker.pick(pet_qty, gift_wrapped)

        # Get vehicles and upload them
        vehicles = []
        while pets:
            vehicle = self.get_vehicle()
            vehicles.append(vehicle)
            self.pet_loader.upload(pets, vehicle)

        # Go
        for vehicle in vehicles:
            vehicle.go(miles)

        # Deliver pets
        for vehicle in vehicles:
            self.pet_loader.download(vehicle)

        # Come back
        for vehicle in vehicles:
            vehicle.come_back()

        # Park
        self.park_vehicles(vehicles)



import logging

from examples.report.cat_example_classes import Cat
from wirinj import Injector, Autowiring, Definitions, Instance

logging.basicConfig(level=logging.DEBUG, format='%(message)s')
inj = Injector(Autowiring(use_singletons=False))
cat = inj.get(Cat)



from unittest import TestCase

from typing import Type
from wirinj.definition import Instance, Definitions
from wirinj import Autowiring
from wirinj.decorators import inject


class TestAutowiring(TestCase):

    def test_full_injection_with_autowiring(self):
        class Cat:
            pass

        class Dog:
            pass

        class Horse:
            pass

        deps = {
            Cat: Instance(),
        }

        @inject(Definitions(deps), Autowiring())
        def fn(cat: Cat, dog: Dog, horse_factory: Type[Horse]):
            horse = horse_factory()

            self.assertIsInstance(cat, Cat)
            self.assertIsInstance(dog, Dog)
            self.assertIsInstance(horse, Horse)

        fn()



from typing import Type

from .classes import Mike, Vehicle, Bob, Car, PetPicker, PetLoader, Pet, Bird, Van, Dog, VehicleBuilder, \
    Truck, Engine, Plate, Wheel, Container, Cat
from wirinj import Definitions, Factory, Singleton, Instance

pet_defs = {
    Dog: Instance(),
    Cat: Instance(),
    Bird: Instance(),

    (Dog, 'sound'): 'Woof',
    (Dog, 'weight'): 10,

    (Cat, 'sound'): 'Meow',
    (Cat, 'weight'): 5,

    (Bird, 'sound'): 'Chirp',
    (Bird, 'weight'): 0.1,
}


vehicle_defs = {
    Engine: Instance(),
    Plate: Instance(),
    Wheel: Instance(),
    Container: Instance(),

    Type[Engine]: Factory(),
    Type[Plate]: Factory(),
    Type[Wheel]: Factory(),
    Type[Container]: Factory(),

    VehicleBuilder: Singleton(),

    (Engine, 'mount_sound'): 'RRRRoarrr',
    (Plate, 'mount_sound'): 'plaf',
    (Wheel, 'mount_sound'): 'pffff',
    (Container, 'mount_sound'): 'BLOOOOM',

    Car: Instance(),
    (Car, 'max_load_weight'): 10,
    (Car, 'recipe'): {
        'engines': 1,
        'plates': 6,
        'wheels': 4,
    },

    Van: Instance(),
    (Van, 'max_load_weight'): 50,
    (Van, 'recipe'): {
        'engines': 1,
        'plates': 8,
        'wheels': 4,
    },

    Truck: Instance(),
    (Truck, 'max_load_weight'): 200,
    (Truck, 'recipe'): {
        'engines': 1,
        'plates': 20,
        'wheels': 12,
        'container': 1,
    },
}

common_defs = {
    PetPicker: Singleton(),
    PetLoader: Singleton(),

    Bob: Singleton(),
    Mike: Singleton(),
}

world_one_defs = {
    (Bob, Vehicle): Singleton(Car),
    (Bob, PetPicker, Type[Pet]): Factory(Bird),

    (Mike, Type[Vehicle]): Factory(Van),
    (Mike, PetPicker, Type[Pet]): Factory(Cat),
}

world_two_defs = {
    (Bob, Vehicle): Singleton(Van),
    (Bob, PetPicker, Type[Pet]): Factory(Cat),

    (Mike, Type[Vehicle]): Factory(Truck),
    (Mike, PetPicker, Type[Pet]): Factory(Dog),
}

world_one = Definitions(
    pet_defs,
    vehicle_defs,
    common_defs,
    world_one_defs,
)

world_two = Definitions(
    pet_defs,
    vehicle_defs,
    common_defs,
    world_two_defs,
)



from examples.report.cat_example_classes import Cat, Head
from wirinj import Injector, Definitions, Instance

inj = Injector(Definitions({
    Cat: Instance(),
    Head: Instance(),
}))
cat2 = inj.get(Cat)



from unittest import TestCase

from wirinj import Autowiring
from wirinj.core import INJECTED
from wirinj.injector import Injector


class Reality(object):
    pass


class Thing:

    reality: Reality = INJECTED

    def __init__(self, param):
        self.param = param


class TestInjector(TestCase):

    def test_field_injection(self):

        inj = Injector(Autowiring())

        thing = inj.get(Thing, 'my-param')
        self.assertIsInstance(thing, Thing)
        self.assertEqual(thing.param, 'my-param')
        self.assertIsInstance(thing.reality, Reality)



from wirinj import inject, Definitions, Instance, Singleton, Factory


class Cat:
    pass


class Dog:
    pass


defs = {
    'sound': 'Meow',
    Cat: Instance(),
    Dog: Singleton(),
    'cat_factory': Factory(Cat),
}


@inject(Definitions(defs))
def fn(cat1: Cat, cat2: Cat, dog: Dog, sound, cat_factory):
    print('sound:', sound)
    print('cat1:', cat1)
    print('cat2:', cat2)
    print('cat1 = cat2:', cat1 is cat2)
    print('dog:', dog)

    cat3 = cat_factory()
    print('cat3:', cat3)


fn()



from abc import ABCMeta, abstractmethod
from inspect import isabstract
from typing import Optional, Sequence

from .definition import Definitions
from .core import Dependency, Arg, Locator, SEPARATOR_OPEN, SEPARATOR_CLOSE, NotSet
from .dependencies import SingletonWrapper, FactoryDependency, InstanceDependency
from .introspect import is_builtin_cls
from .tools import is_typing_type, get_typing_args, is_typing_clause


class AutowiringReportBase(metaclass=ABCMeta):

    @abstractmethod
    def add(self, line):
        pass

    @abstractmethod
    def get(self):
        pass


class NullAutowiringReport(AutowiringReportBase):

    def add(self, line):
        pass

    def get(self):
        return ''


class AutowiringReport(AutowiringReportBase):

    def __init__(self):
        self.lines = []

    def add(self, line):
        self.lines.append(line)

    def get(self):
        if not self.lines:
            return ''

        lines = ''
        for line in self.lines:
            lines += '    {},\n'.format(line)
        return '{0}\n{1}\n\n{2}({{\n{3}}}),\n{4}'.format(
            SEPARATOR_OPEN,
            'Autowiring report:',
            Definitions.__name__,
            lines,
            SEPARATOR_CLOSE,
        )


def is_autowireable_cls(cls):
    return not isabstract(cls) and not is_builtin_cls(cls) and not is_typing_clause(cls)


class Autowiring(Locator):

    def __init__(self, report: Optional[AutowiringReport] = None, use_singletons=True):
        self.report = report if report else NullAutowiringReport()
        self.singletons = {}
        self.use_singletons = use_singletons

    def initialize(self, injector):
        self.injector = injector

    def get(self, creation_path: Sequence[Arg]) -> Optional[Dependency]:
        arg = creation_path[-1]

        if arg.cls is NotSet:
            return None

        try:
            return self.singletons[arg.cls]
        except KeyError:
            pass

        # Something annotated with Type[] is a factory
        if is_typing_type(arg.cls):
            type_cls = get_typing_args(arg.cls)
            if not is_autowireable_cls(type_cls):
                return None
            self.report.add('Type[{}]: Factory()'.format(type_cls.__name__))
            factory_dep = SingletonWrapper(FactoryDependency(type_cls, self.injector))
            self.singletons[arg.cls] = factory_dep
            return factory_dep

        # If is a valid class to be autowired
        elif is_autowireable_cls(arg.cls):

            # With a name, this comes probably from a function arg and therefore it is presumably a singleton
            if arg.name and self.use_singletons:
                self.report.add('{}: Singleton()'.format(arg.cls.__name__))
                singleton_dep = SingletonWrapper(InstanceDependency(arg.cls))
                self.singletons[arg.cls] = singleton_dep
                return singleton_dep

            # Without a name, this comes probably from a factory call and therefore it is presumably an instance
            else:
                self.report.add('{}: Instance()'.format(arg.cls.__name__))
                return InstanceDependency(arg.cls)
        return None



from typing import Sequence, Optional

from .core import Locator, Arg, Dependency


class LocatorChain(Locator):
    def __init__(self, *locator_list: Locator):
        self.locator_list = locator_list

    def initialize(self, injector):
        for finder in self.locator_list:
            finder.initialize(injector)

    def get(self, creation_path: Sequence[Arg]):
        for finder in self.locator_list:
            result = finder.get(creation_path)
            if result is not None:
                return result
        return None


class LocatorCache(Locator):
    def __init__(self, locator: Locator):
        self.real_locator = locator
        self.cache = {}

    def initialize(self, injector):
        self.real_locator.initialize(injector)

    def get(self, creation_path: Sequence[Arg]) -> Optional[Dependency]:

        key = hash(creation_path)

        try:
            return self.cache[key]

        except KeyError as ex:
            result = self.real_locator.get(creation_path)
            self.cache[key] = result
            return result



from typing import Type

from wirinj import INJECTED, inject, Autowiring, Definitions


class MyService:
    def __str__(self):
        return f'<MyService>'


class MyObject:
    my_service: MyService = INJECTED
    my_config: str = INJECTED

    def __str__(self):
        return f'<MyObject> -> my_config: "{self.my_config}"' \
               f', param: {self.param}, my_service: {self.my_service}'

    def __init__(self, param):
        self.param = param


config = {
    'my_config': 'some conf',
}


# The decorator will inject a dependency into each function parameter
# based on its name and/or annotation type
@inject(Definitions(config), Autowiring())
def do(
        my_service: MyService,
        my_object_factory: Type[MyObject]
):
    print(my_service)

    my_object1 = my_object_factory(10)
    print(my_object1)


# Inject and run it
do()



from abc import abstractmethod, ABCMeta
from logging import getLogger
from typing import Optional, Union, Sequence, Any, TypeVar

import wirinj
from .tools import get_cls_name

logger = getLogger(wirinj.__name__)

SEPARATOR_OPEN = '--------------- ' + wirinj.__name__ + ' ---------------'
SEPARATOR_CLOSE = '--------------------------------------'

DEPS_METHOD = '__deps__'
DEPENDENCIES_ARG = '_dependencies'

QUERY_WRAPPED_METHOD = '_query_wrapped_method'

USE_SUBCLASSING_FACTORY = True


class NotSetType(type):
    def __str__(self):
        return 'NotSet'


class NotSet(metaclass=NotSetType):
    pass


class FunctionArgs:
    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs


INJECTED = TypeVar('Injected')


InjectionClauses = [INJECTED]

class Arg:
    def __init__(self, name: Optional[str], cls: Union[NotSetType, Any] = NotSet,
                 default: Union[NotSetType, Any] = NotSet):
        """
        Arg is the representation of a function argument in the context of dependency injection.
        @param name: Name of the argument.
        @param cls: Type annotation of the argument.
        @param default: Default value for the argument.
        """
        self.name = name
        self.cls = cls
        self.default = default

    def __str__(self) -> str:
        return "{}{}{}".format(
            str('' if self.name is None else self.name),
            '' if self.cls is NotSet else ':{}'.format(get_cls_name(self.cls)),
            '' if self.default is NotSet else '={}'.format(str(self.default)),
        )

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return hash((hash(self.name), hash(self.cls), hash(self.default)))


class Dependency(metaclass=ABCMeta):

    def get_class(self) -> Union[Any, NotSet]:
        return NotSet

    def get_dependencies(self) -> Optional[Sequence[Arg]]:
        return None

    @abstractmethod
    def get_instance(self, instance_args: FunctionArgs = None, **deps):
        pass


class Locator(metaclass=ABCMeta):

    def initialize(self, injector):
        pass

    @abstractmethod
    def get(self, creation_path: Sequence[Arg]) -> Optional[Dependency]:
        pass


def filter_direct_args(arg_list: Sequence[Arg], args, kwargs):
    result = []
    i = -1
    for arg in arg_list:

        i += 1

        if i < len(args):
            continue

        if arg.name in kwargs and arg.default != kwargs[arg.name]:
            continue

        result.append(arg)

    return result


def join_arg_sequences(*arg_sequences: Sequence[Arg]):
    result = []
    index = {}
    for arg_sequence in reversed(arg_sequences):
        for arg in arg_sequence:
            if index.get(arg.name):
                continue
            index[arg.name] = True
            result += arg
    return result



from typing import Type

from wirinj import Instance, Factory, Definitions, inject, INJECTED


class Cat:
    sound: str = INJECTED
    weight: float = INJECTED


config = {
    'sound': 'meow',
    'weight': 5,
}

wiring = {
    Cat: Instance(),
    Type[Cat]: Factory()
}


@inject(Definitions(config, wiring))
def fn(cat_factory: Type[Cat]):
    cat = cat_factory()

    print(f'Cat: {cat.sound}, {cat.weight}')


fn()



import logging
from typing import Type

from examples.pet_delivery.classes import Bob, Vehicle, PetPicker, Pet, Mike, Van, Cat, Truck, Dog
from examples.pet_delivery.defs import pet_defs, vehicle_defs, common_defs
from wirinj import Singleton, Factory, Definitions, inject

world_two_defs = {
    (Bob, Vehicle): Singleton(Van),
    (Bob, PetPicker, Type[Pet]): Factory(Cat),

    (Mike, Type[Vehicle]): Factory(Truck),
    (Mike, PetPicker, Type[Pet]): Factory(Dog),
}

world_two = Definitions(
    pet_defs,
    vehicle_defs,
    common_defs,
    world_two_defs,
)

logging.basicConfig(format='%(message)s', level=logging.INFO)


@inject(world_two)
def do(bob: Bob, mike: Mike):
    bob.deliver(100, 5, False)
    bob.deliver(50, 200, True)

    mike.deliver(20, 1000, True)


do()



class Nail:
    pass


class Leg:

    def __init__(self, nail1: Nail, nail2: Nail, nail3: Nail, nail4: Nail, nail5: Nail):
        pass


class Mouth:
    pass


class Ear:
    pass


class Eye:
    pass


class Head:
    def __init__(self, mouth: Mouth, ear1: Ear, ear2: Ear, eye1: Eye, eye2: Eye):
        pass


class Body:
    pass


class Tail:
    pass


class Cat:
    def __init__(self, head: Head, body: Body, tail: Tail, leg1: Leg, leg2: Leg, leg3: Leg, leg4: Leg):
        pass



from typing import Union, Any, Optional, Sequence, Type, Callable

from .core import Dependency, NotSet, Arg, FunctionArgs, USE_SUBCLASSING_FACTORY
from .injector import Injector
from .introspect import get_class_dependencies, instantiate_class, \
    get_func_args, get_func_result
from .tools import get_subclassing_factory, get_func_factory


class ValueDependency(Dependency):

    def __init__(self, value):
        self.value = value

    def get_instance(self, instance_args=None, **deps):
        return self.value


class InstanceDependency(Dependency):

    def __init__(self, cls):
        self.cls = cls

    def get_class(self) -> Union[Any, NotSet]:
        return self.cls

    def get_dependencies(self) -> Optional[Sequence[Arg]]:
        return get_class_dependencies(self.cls)

    def get_instance(self, instance_args=None, **deps):
        return instantiate_class(self.cls, instance_args, **deps)


class SingletonWrapper(Dependency):

    def __init__(self, dependency: Dependency):
        self.dependency = dependency
        self.instance = None

    def get_class(self) -> Union[Any, NotSet]:
        return self.dependency.get_class()

    def get_dependencies(self) -> Optional[Sequence[Arg]]:
        if self.instance:
            return ()
        return self.dependency.get_dependencies()

    def get_instance(self, instance_args=None, **deps):
        if self.instance:
            assert instance_args is None or not instance_args.args and not instance_args.kwargs
        else:
            self.instance = self.dependency.get_instance(instance_args, **deps)
        return self.instance


class FactoryDependency(Dependency):

    def __init__(self, cls, injector: Injector):
        self.cls = cls
        self.injector = injector

    def get_class(self) -> Union[Any, NotSet]:
        return Type[self.cls]

    def get_dependencies(self) -> Optional[Sequence[Arg]]:
        return []

    def get_instance(self, instance_args=None, **deps):
        if USE_SUBCLASSING_FACTORY:
            factory = get_subclassing_factory(self.cls, self.injector.get)
        else:
            factory = get_func_factory(self.cls, self.injector.get)

        return factory


class CustomInstanceDependency(Dependency):

    def __init__(self, func: Callable, cls=None):
        self.func = func
        self.cls = cls

    def get_class(self) -> Union[Any, NotSet]:
        if self.cls:
            return self.cls

        return get_func_result(self.func)

    def get_dependencies(self) -> Optional[Sequence[Arg]]:
        args = get_func_args(self.func)
        return args

    def get_instance(self, instance_args: FunctionArgs = None, **deps):

        if instance_args:
            return self.func(*instance_args.args, **{**deps, **instance_args.kwargs})
        else:
            return self.func(**deps)
